unit SDFilesystemCtrls;

interface

uses
  ComCtrls, Classes, Contnrs, Controls, Windows,
  SysUtils,
  SDUGeneral,
  SDUComCtrls,
  SDFilesystem;

type
{$M+}  // Required to get rid of compiler warning "W1055 PUBLISHED caused RTTI ($M+) to be added to type '%s'"

  TNodeRec = record
    PlaceHolder: boolean;

    Name: string;

    ContentsLoaded: boolean;
    Contents: TSDDirItemList;

  end;
  PNodeRec = ^TNodeRec;

  // Forward declarations
  TSDLoadDirThread = class;
  TSDCustomFilesystemListView = class;


  TSDLoadDirThreadCallback = procedure (thread: TSDLoadDirThread) of object;

  TSDLoadDirThread = class(TThread)
  private
    FFilesystem: TSDCustomFilesystem;
    Paths: TStringList;
    procedure SyncMethod();
  protected
    procedure Execute(); override;
  public
    LoadedPath: string;
    LoadedContents: TSDDirItemList;
    Callback: TSDLoadDirThreadCallback;

    destructor Destroy(); override;
    procedure  AfterConstruction(); override;

    procedure AddPath(pathToLoad: string);

  published
    property Filesystem: TSDCustomFilesystem read FFilesystem write FFilesystem;

  end;
  

  TSDCustomFilesystemTreeView = class(TTreeView)
  private
    FFilesystem: TSDCustomFilesystem;
    FFilesystemListView: TSDCustomFilesystemListView;
    FFolderImgIdxClosed: integer;
    FFolderImgIdxOpen: integer;
    FFolderImage: TImageList;

    FCursorStack: TList;

    procedure AddRootNode();

    procedure AddPlaceHolderNode(Node: TTreeNode);
    procedure DeletePlaceHolderNodesFrom(Node: TTreeNode);
    function  NodeHasPlaceHolderNode(Node: TTreeNode): boolean;

    procedure CreateChildNodes(Node: TTreeNode);
    procedure DeleteAllNodes();

    procedure LoadContents(Node: TTreeNode);

    function  GetNodeForPath(path: string; farAsPossible: boolean): TTreeNode;

    procedure AddExpandedLeafNodes(Node: TTreeNode; expandedLeafNodes: TStringList);
  protected
    function  CanExpand(Node: TTreeNode): boolean; override;
    procedure Delete(Node: TTreeNode); override;
    procedure Change(Node: TTreeNode); override;

    procedure ThreadLoadededContents(thread: TSDLoadDirThread);

    procedure SetFilesystem(newFilesystem: TSDCustomFilesystem);
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy(); override;

    procedure Initialize();

    function  PathToNode(Node: TTreeNode): string;
    function  GoToPath(path: string; farAsPossible: boolean): TTreeNode;

    procedure RefreshNodes();

    procedure BeginUpdate();
    procedure EndUpdate();

  published
    property Filesystem: TSDCustomFilesystem read FFilesystem write SetFilesystem;
    property FilesystemListView: TSDCustomFilesystemListView read FFilesystemListView write FFilesystemListView;

  end;

  TSDFilesystemTreeView = class(TSDCustomFilesystemTreeView)
  end;

  TSDCustomFilesystemListView = class(TSDListView)
  private
    FFilesystem: TSDCustomFilesystem;
    FFilesystemTreeView: TSDCustomFilesystemTreeView;
    FPath: string;
    FContents: TSDDirItemList;
    FShowParentDir: boolean;

    FFileExtnToIconMap_Small: TStringList;
    FFileExtnToIconMap_Large: TStringList;
    FIconsLarge: TImageList;
    FIconsSmall: TImageList;
    FColumnFilename: TListColumn;
    FColumnSize: TListColumn;
    FColumnFileType: TListColumn;
    FColumnTimestamp: TListColumn;
    FColumnAttributes: TListColumn;

    FCursorStack: TList;

  protected
    procedure DblClick(); override;

    function  LoadContents(path: string): boolean;
    procedure RepopulateDisplay();
    procedure SyncIcons();

    procedure GetIconsForFile(filename: string; out smallIconIdx: integer; out largeIconIdx: integer);

    function  GetDirItemByListIdx(idx: integer): TSDDirItem;
    function  GetDirItemSelected(): TSDDirItem;

    procedure SetFilesystem(newFilesystem: TSDCustomFilesystem);

  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy(); override;

    procedure Initialize();

    procedure SetPath(path: string);

    procedure _ColumnSortCompare(Sender: TObject; Item1, Item2: TListItem; Data: Integer; var Compare: Integer); override;

    property DirItem[idx: integer]: TSDDirItem read GetDirItemByListIdx;

    procedure BeginUpdate();
    procedure EndUpdate();

  published
    property Filesystem: TSDCustomFilesystem read FFilesystem write SetFilesystem;
    property FilesystemTreeView: TSDCustomFilesystemTreeView read FFilesystemTreeView write FFilesystemTreeView;
    property Path: string read FPath write SetPath;
    property ShowParentDir: boolean read FShowParentDir write FShowParentDir default FALSE;

    property DirItemSelected: TSDDirItem read GetDirItemSelected;

  end;

  TSDFilesystemListView = class(TSDCustomFilesystemListView)
  end;

procedure Register;

implementation

uses
  Messages, Graphics, Math,
  SDFilesystem_FAT,
  SDUi18n,
  SDUGraphics;

{$IFDEF _NEVER_DEFINED}
// This is just a dummy const to fool dxGetText when extracting message
// information
// This const is never used; it's #ifdef'd out - SDUCRLF in the code refers to
// picks up SDUGeneral.SDUCRLF
const
  SDUCRLF = ''#13#10;
{$ENDIF}

const
  PLACEHOLDER = '<placeholder node>';  // DO NOT TRANSLATE! Just a placeholder marker
  PATH_SEPARATOR = '\';


procedure Register;
begin
  RegisterComponents('SDeanFilesystem', [TSDFilesystemTreeView]);
  RegisterComponents('SDeanFilesystem', [TSDFilesystemListView]);
end;


// farAsPossible - If the path isn't available, go as far along it as possible
function TSDCustomFilesystemTreeView.GoToPath(path: string; farAsPossible: boolean): TTreeNode;
var
  targetNode: TTreeNode;
  tmpNode: TTreeNode;
begin
  BeginUpdate();
  try
    targetNode := GetNodeForPath(path, farAsPossible);
    if (targetNode <> nil) then
      begin
      tmpNode := targetNode.Parent;
      while (tmpNode <> nil) do
        begin
        tmpNode.Expand(FALSE);
        tmpNode := tmpNode.Parent;
        end;

      targetNode.Selected := TRUE;
      targetNode.Focused := TRUE;
      end;

  finally
    EndUpdate();
  end;

  Result := targetNode;
end;

procedure TSDCustomFilesystemTreeView.BeginUpdate();
begin
  FCursorStack.Add(Pointer(self.Cursor));
  self.Cursor := crHourglass;
  Perform(WM_SETCURSOR, Handle, HTCLIENT);  // Force cursor display to update
  Items.BeginUpdate();
end;

procedure TSDCustomFilesystemTreeView.EndUpdate();
begin
  if (FCursorStack.Count > 0) then
    begin
    self.Cursor := TCursor(FCursorStack[FCursorStack.Count-1]);
    Perform(WM_SETCURSOR, Handle, HTCLIENT);  // Force cursor display to update
    FCursorStack.Delete(FCursorStack.Count - 1);
    end;
    
  Items.EndUpdate();
end;

// farAsPossible - If the path isn't available, go as far along it as possible
function TSDCustomFilesystemTreeView.GetNodeForPath(path: string; farAsPossible: boolean): TTreeNode;
var
  retval: TTreeNode;
  pathComponents: TStringList;
  i: integer;
  j: integer;
  lastNode: TTreeNode;
  checkNode: TTreeNode;
  checkNodeRec: PNodeRec;
  allOK: boolean;
  pathPartNode: TTreeNode;
  stlTmp: TStringList;
  pathFirst: string;
  pathRest: string;
  currNodeName: string;
begin
  retval := nil;
  allOK := TRUE;

  // Normalise path
  path := Trim(path);
  path := StringReplace(path, '/', PATH_SEPARATOR, [rfReplaceAll]);
  path := StringReplace(path, '\', PATH_SEPARATOR, [rfReplaceAll]);

  if not(Filesystem.CaseSensitive) then
    begin
    path := uppercase(path);
    end;

  pathComponents := TStringList.Create();
  try
    // Sanity check; a path was passed in, right?
    if allOK then
      begin
      if (length(path) <= 0) then
        begin
        allOK := FALSE;
        end;
      end;

    // Sanity check; the path started with a PATH_SEPARATOR, right?
    if allOK then
      begin
      if (path[1] <> PATH_SEPARATOR) then
        begin
        allOK := FALSE;
        end;
      end;

    // Split path into TStringList
    if allOK then
      begin
      // Remove leading PATH_SEPARATOR; otherwise this would cause the first
      // item taken off with SDUSplitString(...) to be a blank
      System.Delete(path, 1, 1);

      while SDUSplitString(path, pathFirst, pathRest, PATH_SEPARATOR) do
        begin
        pathComponents.Add(pathFirst);
        path := pathRest;
        end;

      // Reinsert the leading "/" we previously took off
      pathComponents.Insert(0, PATH_SEPARATOR);
      end;


    // Handle any "." and ".." in path
    if allOK then
      begin
      stlTmp := TStringList.Create();
      try
        for i:=0 to (pathComponents.Count - 1) do
          begin
          if (pathComponents[i] = DIR_CURRENT_DIR) then
            begin
            // Do nothing - skip
            end
          else if (pathComponents[i] = DIR_PARENT_DIR) then
            begin
            if (stlTmp.Count > 0) then
              begin
              stlTmp.Delete(stlTmp.count-1);
              end
            else
              begin
              allOK := FALSE;
              break;
              end;
            end
          else
            begin
            stlTmp.Add(pathComponents[i]);
            end;

          end;

        pathComponents.Assign(stlTmp);
      finally
        stlTmp.Free();
      end;
      
      end;


    // Seek node...
    if allOK then
      begin
      // Root node is always the root node
      lastNode := self.Items.GetFirstNode();
      for i:=1 to (pathComponents.Count - 1) do
        begin
        // Create nodes as we go, if required...
        if NodeHasPlaceHolderNode(lastNode) then
          begin
          CreateChildNodes(lastNode);
          end;

        // Check all of the current node's children to try and find the next
        // path element
        pathPartNode := nil;
        for j:=0 to (lastNode.Count - 1) do
          begin
          checkNode := lastNode.Item[j];
          checkNodeRec := checkNode.Data;

          // (We're case insensitive (and below))
          currNodeName := checkNodeRec.Name;
          if not(Filesystem.CaseSensitive) then
            begin
            currNodeName := uppercase(currNodeName);
            end;

          if (currNodeName = pathComponents[i]) then
            begin
            pathPartNode := checkNode;
            break;
            end;
            
          end;

        if (pathPartNode = nil) then
          begin
          allOK := FALSE;
          break;
          end;

        lastNode := pathPartNode;
        end;

      retval := lastNode;
      end;

    // If there was a problem, make sure we return NIL
    if (
        not(farAsPossible) and
        not(allOK)
       ) then
      begin
      retval := nil;
      end;

  finally
    pathComponents.Free();
  end;

  Result := retval;
end;


constructor TSDCustomFilesystemTreeView.Create(AOwner: TComponent);
begin
  inherited;

  FCursorStack := TList.Create();

  // Prevent user from changing node captions
  self.Readonly := TRUE;

  // Leave visual indicator of selected node when control not focussed
  self.HideSelection := FALSE;

  // No line to the root node
  self.ShowRoot := FALSE;

  FFolderImage:= TImageList.Create(nil);
  FFolderImage.BlendColor := self.Color;
  FFolderImage.BkColor := self.Color;

  // Get icon file and icon index within icon file
  FFolderImgIdxClosed := SDULoadDLLIconToList(
                                              DLL_SHELL32,
                                              TRUE,
                                              DLL_SHELL32_FOLDER_CLOSED,
                                              FFolderImage
                                             );
  FFolderImgIdxOpen   := SDULoadDLLIconToList(
                                              DLL_SHELL32,
                                              TRUE,
                                              DLL_SHELL32_FOLDER_OPEN,
                                              FFolderImage
                                             );

end;

destructor  TSDCustomFilesystemTreeView.Destroy();
begin
  FFolderImage.Free();
  FCursorStack.Free();
  inherited;
end;

procedure TSDCustomFilesystemTreeView.Initialize();
begin
  self.Images := FFolderImage;
  DeleteAllNodes();
  AddRootNode();
end;

procedure TSDCustomFilesystemTreeView.SetFilesystem(newFilesystem: TSDCustomFilesystem);
begin
  FFilesystem := newFilesystem;
  if (Filesystem = nil) then
    begin
    DeleteAllNodes();
    end;
end;

procedure TSDCustomFilesystemTreeView.AddRootNode();
var
  newNode: TTreeNode;
  newNodeRec: PNodeRec;
begin
  // Create initial root node
  newNodeRec := new(PNodeRec);
  newNodeRec.PlaceHolder := FALSE;
  newNodeRec.Name := PATH_SEPARATOR;
  newNodeRec.ContentsLoaded := FALSE;
  newNodeRec.Contents := nil;
  newNode:= self.Items.AddFirst(nil, newNodeRec.Name);
  newNode.Data := newNodeRec;

  AddPlaceHolderNode(newNode);

  newNode.Selected := TRUE;
  newNode.Focused := TRUE;
end;

procedure TSDCustomFilesystemTreeView.AddPlaceHolderNode(Node: TTreeNode);
var
  newNode: TTreeNode;
  newNodeRec: PNodeRec;
begin
  newNodeRec := new(PNodeRec);
  newNodeRec.PlaceHolder := TRUE;
  newNodeRec.Name := PLACEHOLDER;
  newNodeRec.ContentsLoaded := FALSE;
  newNode := self.Items.AddChild(Node, newNodeRec.Name);
  newNode.Data := newNodeRec;
end;

// Delete all place holder nodes from the specified node
procedure TSDCustomFilesystemTreeView.DeletePlaceHolderNodesFrom(Node: TTreeNode);
var
  currNode: TTreeNode;
  currNodeRec: PNodeRec;
  i: integer;
begin
  for i:=(Node.Count - 1) downto 0 do
    begin
    currNode := Node.Item[i];
    currNodeRec := currNode.Data;
    if currNodeRec.PlaceHolder then
      begin
      Node.Item[i].Delete();
      end;

    end;

end;

function TSDCustomFilesystemTreeView.NodeHasPlaceHolderNode(Node: TTreeNode): boolean;
var
  currNode: TTreeNode;
  currNodeRec: PNodeRec;
  i: integer;
  retval: boolean;
begin
  retval := FALSE;

  for i:=(Node.Count - 1) downto 0 do
    begin
    currNode := Node.Item[i];
    currNodeRec := currNode.Data;
    if currNodeRec.PlaceHolder then
      begin
      retval := TRUE;
      break
      end;

    end;

  Result := retval;
end;

procedure TSDCustomFilesystemTreeView.DeleteAllNodes();
begin
  self.Items.Clear();
end;

procedure TSDCustomFilesystemTreeView.CreateChildNodes(Node: TTreeNode);
var
  newNode: TTreeNode;
  nodeRec: PNodeRec;
  newNodeRec: PNodeRec;
  currName: string;
  i: integer;
  nodePath: string;
begin
  nodeRec := Node.Data;

  LoadContents(Node);
  if nodeRec.ContentsLoaded then
    begin
    DeletePlaceHolderNodesFrom(Node);

    nodePath := PathToNode(Node);

    for i:=0 to (nodeRec.Contents.Count - 1) do
      begin
      currName := nodeRec.Contents[i].Filename;

      if (
          nodeRec.Contents[i].IsDirectory and
          (nodeRec.Contents[i].Filename <> DIR_CURRENT_DIR) and
          (nodeRec.Contents[i].Filename <> DIR_PARENT_DIR)
         ) then
        begin
        newNodeRec := new(PNodeRec);
        newNodeRec.PlaceHolder := FALSE;
        newNodeRec.Name := currName;
        newNodeRec.ContentsLoaded := FALSE;
        newNodeRec.Contents := nil;
        newNode:= self.Items.AddChild(Node, newNodeRec.Name);
        newNode.Data := newNodeRec;
        newNode.ImageIndex := FFolderImgIdxClosed;
        newNode.SelectedIndex:= FFolderImgIdxOpen;

        AddPlaceHolderNode(newNode);
        end;

      end;

    // Get the nodes just added into alphabetical order
    Node.AlphaSort(FALSE);
    end;

end;

procedure TSDCustomFilesystemTreeView.ThreadLoadededContents(thread: TSDLoadDirThread);
var
  node: TTreeNode;
  nodeRec: PNodeRec;
begin
  node := GetNodeForPath(thread.LoadedPath, FALSE);
  if (node <> nil) then
    begin
    nodeRec := Node.Data;

    if not(nodeRec.ContentsLoaded) then
      begin
      nodeRec.Contents := TSDDirItemList.Create();
      nodeRec.Contents.Assign(thread.LoadedContents);
      nodeRec.ContentsLoaded := TRUE;

      if NodeHasPlaceHolderNode(node) then
        begin
        CreateChildNodes(node);
        end;

      end;

    end;

end;

procedure TSDCustomFilesystemTreeView.LoadContents(Node: TTreeNode);
var
  nodeRec: PNodeRec;
  tmpDirItemList: TSDDirItemList;
begin
  noderec := Node.Data;
  if not(nodeRec.ContentsLoaded) then
    begin
    tmpDirItemList := TSDDirItemList.Create();
    try
      nodeRec.ContentsLoaded := FFileSystem.LoadContentsFromDisk(PathToNode(Node), tmpDirItemList);

      if nodeRec.ContentsLoaded then
        begin
        nodeRec.Contents := tmpDirItemList;
        end
      else
        begin
        // If there was a problem, free off dir list object created
        tmpDirItemList.Free();
        end;
        
    except
      on E:Exception do
        begin
        tmpDirItemList.Free();
        end;
    end;

    end;

end;

function TSDCustomFilesystemTreeView.PathToNode(Node: TTreeNode): string;
var
  retval: string;
  currNode: TTreeNode;
  currNodeRec: PNodeRec;
begin
  retval := '';

  currNode := Node;
  while (currNode <> nil) do
    begin
    currNodeRec := currNode.Data;

    if (retval = '') then
      begin
      retval := currNodeRec.Name;
      end
    else
      begin
      if (currNode.Parent = nil) then
        begin
        retval := PATH_SEPARATOR + retval;
        end
      else
        begin
        retval := currNodeRec.Name + PATH_SEPARATOR + retval;
        end;
      end;
      
    currNode := currNode.Parent;             
    end;

  Result := retval;
end;

procedure TSDCustomFilesystemTreeView.AddExpandedLeafNodes(Node: TTreeNode; expandedLeafNodes: TStringList);
var
  i: integer;
  preChildCnt: integer;
begin
  // Add all expanded leaf nodes to expandedLeafNodes (place holder nodes not
  // considered leaf nodes)

  if Node.Expanded then
    begin
    preChildCnt := expandedLeafNodes.Count;

    for i:=0 to (Node.Count - 1) do
      begin
      AddExpandedLeafNodes(Node.Item[i], expandedLeafNodes);
      end;

    // If none of the child nodes where expanded, add this node on
    if (preChildCnt = expandedLeafNodes.Count) then
      begin
      expandedLeafNodes.Add(PathToNode(Node));
      end;
    end;

end;

function TSDCustomFilesystemTreeView.CanExpand(Node: TTreeNode): boolean;
var
  i: integer;
  currNode: TTreeNode;
  currThread: TSDLoadDirThread;
  retval: boolean;
begin
  retval := inherited CanExpand(Node);

  if retval then
    begin
    BeginUpdate();
    try
      if NodeHasPlaceHolderNode(Node) then
        begin
        CreateChildNodes(Node);
        end;

      retval := Node.HasChildren;

      currThread := TSDLoadDirThread.Create(TRUE);
      for i:=0 to (Node.Count - 1) do
        begin
        currNode := Node.Item[i];
        currThread.AddPath(PathToNode(currNode));
        end;
      currThread.Filesystem := self.Filesystem;
      currThread.Callback := ThreadLoadededContents;
      currThread.FreeOnTerminate := TRUE;
      // Put the priority of the thread down; the application gets sluggish
      // otherwise
      currThread.Priority := tpLower;
      currThread.Resume();
    finally
      EndUpdate();
    end;

    end;

  Result := retval;
end;

procedure TSDCustomFilesystemTreeView.Delete(Node: TTreeNode);
var
  nodeRec: PNodeRec;
begin
  inherited;

  nodeRec := Node.Data;

  if not(nodeRec.PlaceHolder) then
    begin
    if nodeRec.ContentsLoaded then
      begin
      nodeRec.ContentsLoaded := FALSE;
      end;

    if (nodeRec.Contents <> nil) then
      begin
      nodeRec.Contents.Free();
      nodeRec.Contents := nil;
      end;

    end;

  Dispose(nodeRec);

end;

procedure TSDCustomFilesystemTreeView.Change(Node: TTreeNode);
begin
  LoadContents(Node);

  if NodeHasPlaceHolderNode(Node) then
    begin
    CreateChildNodes(Node);
    end;

  if Assigned(FFilesystemListView) then
    begin
    FFilesystemListView.SetPath(PathToNode(Node));
    end;

  inherited;

end;

procedure TSDCustomFilesystemTreeView.RefreshNodes();
var
  expandedLeafNodes: TStringList;
  i: integer;
  currNode: TTreeNode;
  lastPath: string;
  lastExpanded: boolean;
begin
  BeginUpdate();
  try
    lastPath := PathToNode(self.Selected);
    lastExpanded := FALSE;
    if (self.Selected <> nil) then
      begin
      lastExpanded := self.Selected.Expanded;
      end;

    expandedLeafNodes:= TStringList.Create();
    try
      // Note: We only ever process the one root node
      AddExpandedLeafNodes(self.Items.GetFirstNode(), expandedLeafNodes);

      DeleteAllNodes();
      AddRootNode();

      for i:=0 to (expandedLeafNodes.count - 1) do
        begin
        currNode := GoToPath(expandedLeafNodes[i], TRUE);
        currNode.Expanded := TRUE;
        end;

    finally
      expandedLeafNodes.Free();
    end;

    currNode := GoToPath(lastPath, TRUE);
    if (lastPath = PathToNode(currNode)) then
      begin
      currNode.Expanded := lastExpanded;
      end
    else
      begin
      currNode.Expanded := TRUE;
      end;

  finally
    EndUpdate();
  end;

end;

procedure TSDCustomFilesystemListView.SetFilesystem(newFilesystem: TSDCustomFilesystem);
begin
  FFilesystem := newFilesystem;
  if (Filesystem = nil) then
    begin
    self.Clear();
    end;
end;

procedure TSDCustomFilesystemListView.BeginUpdate();
begin
  FCursorStack.Add(Pointer(self.Cursor));
  self.Cursor := crHourglass;
  Perform(WM_SETCURSOR, Handle, HTCLIENT);  // Force cursor display to update
  Items.BeginUpdate();
end;

procedure TSDCustomFilesystemListView.EndUpdate();
begin
  if (FCursorStack.Count > 0) then
    begin
    self.Cursor := TCursor(FCursorStack[FCursorStack.Count-1]);
    Perform(WM_SETCURSOR, Handle, HTCLIENT);  // Force cursor display to update
    FCursorStack.Delete(FCursorStack.Count - 1);
    end;
    
  Items.EndUpdate();
end;

procedure TSDCustomFilesystemListView.SetPath(path: string);
begin
  BeginUpdate();
  try
    self.Clear();  // Clear the display; otherwise when self.Clear() is called
                   // in RepopulateDisplay(...) (below), it can fire off the
                   // OnChanged event - at which point the display will be out
                   // of sync with FContents - and if the OnChange event
                   // accesses FContents via DirItem[...] based on self.Count,
                   // it can crash the DirItem[...] caller
    try
      if LoadContents(path) then
        begin
        FPath := path;
        end;
    finally
      RepopulateDisplay();
    end;

  finally
    EndUpdate();
  end;
  
end;

// Update icons for display
// Needed as large icon indexes may not match small icon indexes
procedure TSDCustomFilesystemListView.SyncIcons();
var
  i: integer;
  dirItem: TSDDirItem;
  iconIdxSmall: integer;
  iconIdxLarge: integer;
begin
  for i:=0 to (Self.Items.Count - 1) do
    begin
    dirItem := GetDirItemByListIdx(i);
    if dirItem.IsDirectory then
      begin
      GetIconsForFile(FILE_TYPE_DIRECTORY, iconIdxSmall, iconIdxLarge);
      end
    else
      begin
      GetIconsForFile(dirItem.Filename, iconIdxSmall, iconIdxLarge);
      end;

    if (self.ViewStyle = vsIcon) then
      begin
      self.Items[i].ImageIndex := iconIdxLarge;
      end
    else
      begin
      self.Items[i].ImageIndex := iconIdxSmall;
      end;
    end;

end;

// Clear and repopulate display with loaded dir content
procedure TSDCustomFilesystemListView.RepopulateDisplay();
var
  i: integer;
  size: string;
  fileType: string;
  sizeKB: integer;
  date: string;
  attributes: string;
  currItem: TSDDirItem;
begin
  BeginUpdate();
  try
    self.Clear();

    for i:=0 to (FContents.Count - 1) do
      begin
      currItem := FContents[i];

      // Skip volume labels, devices, and "."
      // Also skip "..", unless ShowParentDir is TRUE
      if not(
             currItem.IsFile or
             (
              // It's a directory, but not the current directory
              currItem.IsDirectory and
              (currItem.Filename <> DIR_CURRENT_DIR) and
              (
               (currItem.Filename <> DIR_PARENT_DIR) or
               (
                (currItem.Filename = DIR_PARENT_DIR) and
                ShowParentDir 
               )
              )
             )
            ) then
        begin
        continue;
        end;

      size := '';
      if currItem.IsDirectory then
        begin
        fileType := SDUGetFileType_Description(FILE_TYPE_DIRECTORY);
        end
      else
        begin
        // Always display size in KB
        sizeKB := (currItem.Size div BYTES_IN_KILOBYTE);
        // Round up, like MS Explorer does...
        if ((currItem.Size mod BYTES_IN_KILOBYTE) > 0) then
          begin
          sizeKB := sizeKB + 1;
          end;
        // Format with thousands separator and KB units
        size := SDUIntToStrThousands(sizeKB) + ' ' + UNITS_STORAGE_KB;

        fileType := SDUGetFileType_Description(currItem.Filename);
        end;

      date := '';
      if (currItem.TimestampLastModified.Date <> 0) then
        begin
        date := DateTimeToStr(TimeStampToDateTime(currItem.TimestampLastModified))
        end;

      attributes := '';
      if currItem.IsReadonly then
        begin
        attributes := attributes + _('R');
        end;
      if (currItem is TSDDirItem_FAT) then
        begin
        if TSDDirItem_FAT(currItem).IsHidden then
          begin
          attributes := attributes + _('H');
          end;
        if TSDDirItem_FAT(currItem).IsSystem then
          begin
          attributes := attributes + _('S');
          end;
        if TSDDirItem_FAT(currItem).IsArchive then
          begin
          attributes := attributes + _('A');
          end;
        end;

      Self.AppendRow(
                     [
                      currItem.Filename,
                      size,
                      fileType,
                      date,
                      attributes
                     ],
                     Pointer(i)
                    );
      end;

    SyncIcons();

    // Sort the items...
  // These two methods of sorting don't work?!
  //  self.SortType := stText;
  //  self.AlphaSort();
    self.Tag := 1;  // Ensure that sort is on filename
    self.ColClick(FColumnFilename);

  finally
    EndUpdate();
  end;

end;

procedure TSDCustomFilesystemListView.DblClick();
var
  contentsIdx: integer;
  target: string;
begin
  inherited;

  if (self.SelCount = 1) then
    begin
    contentsIdx := integer(self.Selected.Data);
    target := IncludeTrailingPathDelimiter(Path) + FContents[contentsIdx].Filename;

    if Assigned(FilesystemTreeView) then
      begin
      FilesystemTreeView.GoToPath(target, TRUE);
      end
    else
      begin
      self.SetPath(target);
      end;

    end;

end;

function TSDCustomFilesystemListView.LoadContents(path: string): boolean;
var
  tmpDirItemList: TSDDirItemList;
  retval: boolean;
begin
  retval := FALSE;

  tmpDirItemList := TSDDirItemList.Create();
  try
    retval := FFileSystem.LoadContentsFromDisk(path, tmpDirItemList);
    if retval then
      begin
      if (FContents <> nil) then
        begin
        FContents.Free();
        end;

      FContents := tmpDirItemList;
      end
    else
      begin
      // If there was a problem, free off dir list object created
      tmpDirItemList.Free();
      end;
        
  except
    on E:Exception do
      begin
      tmpDirItemList.Free();
      end;
  end;

  Result := retval;
end;

constructor TSDCustomFilesystemListView.Create(AOwner: TComponent);
begin
  inherited;

  // Prevent user from changing filenames
  self.Readonly := TRUE;

  FCursorStack := TList.Create();

  FShowParentDir := FALSE;

  FColumnFilename:= nil;
  FColumnSize:= nil;
  FColumnFileType:= nil;
  FColumnTimestamp:= nil;
  FColumnAttributes:= nil;

  FContents := nil;

  FFileExtnToIconMap_Large:= TStringList.Create();
  FIconsLarge:= TImageList.Create(nil);
  FIconsLarge.Width  := 32;
  FIconsLarge.Height := 32;
  FIconsLarge.BkColor := self.Color;

  FFileExtnToIconMap_Small:= TStringList.Create();
  FIconsSmall:= TImageList.Create(nil);
  FIconsSmall.Width  := 16;
  FIconsSmall.Height := 16;
  FIconsSmall.BkColor := self.Color;

end;

destructor TSDCustomFilesystemListView.Destroy();
begin
  FFileExtnToIconMap_Large.Free();
  FIconsLarge.Free();
  FFileExtnToIconMap_Small.Free();
  FIconsSmall.Free();

  FCursorStack.Free();

  inherited;
end;

procedure TSDCustomFilesystemListView.Initialize();
begin
  self.Items.Clear();
  self.RowSelect := TRUE;
  self.Columns.Clear();

  FColumnFilename         := self.AddCol(_('Name'));
  // FColumnFilename.Width set below
  FColumnSize             := self.AddCol(_('Size'));
  FColumnSize.Width       := 75;
  FColumnFileType         := self.AddCol(_('Type'));
  FColumnFileType.Width   := 150;
  FColumnTimestamp        := self.AddCol(_('Date modified'));
  FColumnTimestamp.Width  := 120;
  FColumnAttributes       := self.AddCol(_('Attributes'));
  FColumnAttributes.Width := 100;

  // -50 so the scrollbar doesn't show when first shown
  FColumnFilename.Width := max(
                               100,
                               (
                                self.Width - (
                                         FColumnSize.width +
                                         FColumnFileType.width +
                                         FColumnTimestamp.width +
                                         FColumnAttributes.width
                                        ) - 50
                               )
                              );

  Self.LargeImages := FIconsLarge;
  Self.SmallImages := FIconsSmall;
end;

procedure TSDCustomFilesystemListView._ColumnSortCompare(Sender: TObject; Item1, Item2: TListItem; Data: Integer; var Compare: Integer);
var
  dirItem1: TSDDirItem;
  dirItem2: TSDDirItem;
begin
  Compare := 0;

  dirItem1:= FContents[integer(Item1.Data)];
  dirItem2:= FContents[integer(Item2.Data)];

  // Parent dir comes first, regardless of anything else...
  if dirItem1.Filename = DIR_PARENT_DIR then
    begin
    Compare := -1;
    end
  else if dirItem2.Filename = DIR_PARENT_DIR then
    begin
    Compare := 1;
    end;

  // Directories come before files...
  if (Compare = 0) then
    begin
    if (dirItem1.IsFile and dirItem2.IsDirectory) then
      begin
      Compare := 1;
      end
    else if (dirItem1.IsDirectory and dirItem2.IsFile) then
      begin
      Compare := -1;
      end;
    end;

  // If the two items haven't got a mandatory sort order, fallback to the
  // default sort order...
  if (Compare = 0) then
    begin
    inherited;
    end;
end;

procedure TSDCustomFilesystemListView.GetIconsForFile(filename: string; out smallIconIdx: integer; out largeIconIdx: integer);
var
  DLLFilename: string;
  DLLIconIdx: integer;
  ucExtn: string;
begin
  // "Special" icons
  if (filename = FILE_TYPE_DIRECTORY) then
    begin
    // Lowercase intentional; this way it won't get confused with the
    // uppercase (*real*) extensions added to the TStringList later
    ucExtn := lowercase(filename);
    end
  else
    begin
    ucExtn := uppercase(ExtractFileExt(filename));
    end;

  smallIconIdx := FFileExtnToIconMap_Small.IndexOf(ucExtn);
  largeIconIdx := FFileExtnToIconMap_Large.IndexOf(ucExtn);

  // If neither icon already cached, get new icons
  if (
      (smallIconIdx < 0) and
      (largeIconIdx < 0)
     ) then
    begin
    // Get icon file and icon index within icon file
    SDUGetFileType_Icon(filename, DLLFilename, DLLIconIdx);

    if (DLLFilename = '') then
      begin
      DLLFilename := DLL_SHELL32;
      DLLIconIdx := DLL_SHELL32_DEFAULT_FILE;
      end;

    if (SDULoadDLLIconToList(DLLFilename, TRUE, DLLIconIdx, FIconsSmall) >= 0) then
      begin
      FFileExtnToIconMap_Small.Add(ucExtn);
      end
    else
      begin
      if (SDULoadDLLIconToList(DLL_SHELL32, TRUE, DLL_SHELL32_DEFAULT_FILE, FIconsSmall) >= 0) then
        begin
        FFileExtnToIconMap_Small.Add(ucExtn);
        end;
      end;

    if (SDULoadDLLIconToList(DLLFilename, FALSE, DLLIconIdx, FIconsLarge) >= 0) then
      begin
      FFileExtnToIconMap_Large.Add(ucExtn);
      end
    else
      begin
      if (SDULoadDLLIconToList(DLL_SHELL32, FALSE, DLL_SHELL32_DEFAULT_FILE, FIconsLarge) >= 0) then
        begin
        FFileExtnToIconMap_Large.Add(ucExtn);
        end;
      end;
    end;

  smallIconIdx := FFileExtnToIconMap_Small.IndexOf(ucExtn);
  largeIconIdx := FFileExtnToIconMap_Large.IndexOf(ucExtn);

end;

function TSDCustomFilesystemListView.GetDirItemByListIdx(idx: integer): TSDDirItem;
begin
  Result := FContents[integer(self.Items[idx].Data)];
end;


function TSDCustomFilesystemListView.GetDirItemSelected(): TSDDirItem;
var
  retval: TSDDirItem;
  idx: integer;
begin
  retval := nil;

  idx := self.SelectedIdx;
  if (idx >= 0) then
    begin
    retval := GetDirItemByListIdx(idx);
    end;

  Result := retval;
end;

procedure TSDLoadDirThread.AfterConstruction();
begin
  inherited;
  Paths := TStringList.Create();
  LoadedContents := TSDDirItemList.Create();
end;

destructor TSDLoadDirThread.Destroy();
begin
  LoadedContents.Free();
  Paths.Free();
  inherited;
end;

procedure TSDLoadDirThread.Execute();
var
  i: integer;
begin
  for i:=0 to (Paths.count - 1) do
    begin
    LoadedContents.Clear();
    if Filesystem.LoadContentsFromDisk(Paths[i], LoadedContents) then
      begin
      LoadedPath := Paths[i];
      Synchronize(SyncMethod);
      end;

    end;
    
end;

procedure TSDLoadDirThread.SyncMethod();
begin
  Callback(self);
end;

procedure TSDLoadDirThread.AddPath(pathToLoad: string);
begin
  Paths.Add(pathToLoad);
end;


END.


